import random
import socket
from pathlib import Path

import click
import youtube_dl


AUDIO_MODELS = [("Google Inc.", "Chromecast Audio")]
ULTRA_MODELS = [("Xiaomi", "MIBOX3"), ("Google Inc.", "Chromecast Ultra")]

BEST_MAX_2K = "best[width <=? 1920][height <=? 1080]"
BEST_MAX_4K = "best[width <=? 3840][height <=? 2160]"
BEST_ONLY_AUDIO = "bestaudio"
BEST_FALLBACK = "/best"
MAX_50FPS = "[fps <=? 50]"
TWITCH_NO_60FPS = "[format_id != 1080p60__source_][format_id != 720p60]"
MIXCLOUD_NO_DASH = "[format_id != dash-a1-x3]"
AUDIO_FORMAT = BEST_ONLY_AUDIO + MIXCLOUD_NO_DASH + BEST_FALLBACK
ULTRA_FORMAT = BEST_MAX_4K
STANDARD_FORMAT = BEST_MAX_2K + MAX_50FPS + TWITCH_NO_60FPS


class CattInfoError(click.ClickException):
    pass


class StreamInfo:
    def __init__(self, video_url, model=None, host=None):
        if "://" not in video_url:
            self._local_file = video_url
            self.local_ip = self._get_local_ip(host)
            self.port = random.randrange(45000, 47000)
            self.is_local_file = True
        else:
            self._ydl = youtube_dl.YoutubeDL({"quiet": True, "no_warnings": True})
            self._preinfo = self._get_stream_preinfo(video_url)
            # Some playlist urls needs to be re-processed (such as youtube channel urls).
            if self._preinfo.get("ie_key"):
                self._preinfo = self._get_stream_preinfo(self._preinfo["url"])
            self.local_ip = None
            self.port = None
            self.is_local_file = False

            if model in AUDIO_MODELS:
                self._best_format = AUDIO_FORMAT
            elif model in ULTRA_MODELS:
                self._best_format = ULTRA_FORMAT
            else:
                self._best_format = STANDARD_FORMAT

            if self.is_playlist:
                self._entries = list(self._preinfo["entries"])
                self._active_entry = None
            else:
                self._info = self._get_stream_info(self._preinfo)

    @property
    def is_video(self):
        return not self.is_local_file and not self.is_playlist

    @property
    def is_playlist(self):
        return not self.is_local_file and "entries" in self._preinfo

    @property
    def extractor(self):
        return self._preinfo["extractor"].split(":")[0] if not self.is_local_file else None

    @property
    def video_title(self):
        if self.is_local_file:
            return Path(self._local_file).name
        elif self.is_playlist:
            return None
        else:
            # "preinfo" does not contain a "title" key, when the user casts
            # an url that points directly to a media file.
            try:
                return self._preinfo["title"]
            except KeyError:
                return self._preinfo["webpage_url_basename"].split(".")[0]

    @property
    def video_url(self):
        if self.is_local_file:
            return "http://%s:%s/?loaded_from_catt" % (self.local_ip, self.port)
        elif self.is_playlist:
            return None
        else:
            return self._get_stream_url(self._info)

    @property
    def video_id(self):
        return self._preinfo["id"] if self.is_video else None

    @property
    def video_thumbnail(self):
        return self._preinfo.get("thumbnail") if self.is_video else None

    @property
    def playlist_length(self):
        return len(self._entries) if self.is_playlist else None

    @property
    def playlist_all_ids(self):
        if self.is_playlist and self._entries and self._entries[0].get("id"):
            return [entry["id"] for entry in self._entries]
        else:
            return None

    @property
    def playlist_title(self):
        return self._preinfo["title"] if self.is_playlist else None

    @property
    def playlist_id(self):
        return self._preinfo["id"] if self.is_playlist else None

    @property
    def playlist_entry_title(self):
        return self._active_entry["title"] if self.is_playlist else None

    @property
    def playlist_entry_url(self):
        if self.is_playlist:
            return self._get_stream_url(self._get_stream_info(self._active_entry))
        else:
            return None

    @property
    def playlist_entry_id(self):
        return self._active_entry["id"] if self.is_playlist else None

    @property
    def playlist_entry_thumbnail(self):
        return self._active_entry.get("thumbnail") if self.is_playlist else None

    def set_playlist_entry(self, number):
        """
        Must be called with valid entry number
        before playlist entry properties can be accessed.
        """

        if self.is_playlist:
            # Some playlist entries needs to be re-processed.
            if self._entries[number].get("ie_key"):
                self._active_entry = self._get_stream_preinfo(self._entries[number]["url"])
            else:
                self._active_entry = self._entries[number]

    def _get_local_ip(self, cc_host):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect((cc_host, 0))
        return sock.getsockname()[0]

    def _get_stream_preinfo(self, video_url):
        try:
            return self._ydl.extract_info(video_url, process=False)
        except youtube_dl.utils.DownloadError:
            raise CattInfoError("Remote resource not found.")

    def _get_stream_info(self, preinfo):
        try:
            return self._ydl.process_ie_result(preinfo, download=False)
        except (youtube_dl.utils.ExtractorError, youtube_dl.utils.DownloadError):
            raise CattInfoError("Youtube-dl extractor failed.")

    def _get_stream_url(self, info):
        format_selector = self._ydl.build_format_selector(self._best_format)

        try:
            best_format = next(format_selector(info))
        except StopIteration:
            raise CattInfoError("No suitable format was found.")
        # This is thrown when url points directly to media file.
        except KeyError:
            best_format = info

        return best_format["url"]



# -*- coding: utf-8 -*-
import configparser
import random
import time
from pathlib import Path
from threading import Thread

import click

from .controllers import (
    Cache,
    CastState,
    get_chromecast,
    get_chromecasts,
    PlaybackError,
    setup_cast,
    StateFileError
)
from .http_server import serve_file
from .util import warning


CONFIG_DIR = Path(click.get_app_dir("catt"))
CONFIG_PATH = Path(CONFIG_DIR, "catt.cfg")
STATE_PATH = Path(CONFIG_DIR, "state.json")


class CattCliError(click.ClickException):
    pass


class CattTimeParamType(click.ParamType):
    def convert(self, value, param, ctx):
        try:
            tdesc = [int(x) for x in value.split(":")]
            tlen = len(tdesc)
            if (tlen > 1 and any(t > 59 for t in tdesc)) or tlen > 3:
                raise ValueError
        except ValueError:
            self.fail("%s is not a valid time description" % value, param, ctx)

        tdesc.reverse()
        return sum(tdesc[p] * 60 ** p for p in range(tlen))


CATT_TIME = CattTimeParamType()


def human_time(seconds):
    return time.strftime("%H:%M:%S", time.gmtime(seconds))


def process_url(ctx, param, value):
    if value == "-":
        stdin_text = click.get_text_stream("stdin")
        if not stdin_text.isatty():
            value = stdin_text.read().strip()
        else:
            raise CattCliError("No input received from stdin.")
    if "://" not in value:
        if ctx.info_name != "cast":
            raise CattCliError("Local file not allowed as argument to this command.")
        if not Path(value).exists():
            raise CattCliError("The chosen file does not exist.")
    return value


def process_path(ctx, param, value):
    path = Path(value) if value else None
    if path and (path.is_dir() or not path.parent.exists()):
        raise CattCliError("The specified path is invalid.")
    return path


def get_device(ctx, param, value):
    try:
        return ctx.default_map["aliases"][value]
    except KeyError:
        return value


@click.group()
@click.option("--delete-cache", is_flag=True,
              help="Empty the Chromecast discovery cache.")
@click.option("-d", "--device", metavar="NAME",
              callback=get_device, help="Select Chromecast device.")
@click.pass_context
def cli(ctx, delete_cache, device):
    if delete_cache:
        Cache().clear()
    ctx.obj["device"] = device


@cli.command(short_help="Write the name of default Chromecast "
                        "device to config file.")
@click.pass_obj
def write_config(settings):
    if settings.get("device"):
        # This is so we fail if the specified Chromecast cannot be found.
        get_chromecast(settings["device"])
        writeconfig(settings)
    else:
        raise CattCliError("No device specified.")


@cli.command(short_help="Send a video to a Chromecast for playing.")
@click.argument("video_url", callback=process_url)
@click.option("-f", "--force-default", is_flag=True,
              help="Force use of the default Chromecast app (use if a custom app doesn't work).")
@click.option("-r", "--random-play", is_flag=True,
              help="Play random item from playlist, if applicable.")
@click.pass_obj
def cast(settings, video_url, force_default, random_play):
    controller = "default" if force_default else None
    cst, stream = setup_cast(settings["device"], video_url=video_url,
                             prep="app", controller=controller)

    if stream.is_local_file:
        click.echo("Casting local file %s..." % video_url)
        click.echo("Playing %s on \"%s\"..." % (stream.video_title, cst.cc_name))

        thr = Thread(target=serve_file,
                     args=(video_url, stream.local_ip, stream.port))

        thr.setDaemon(True)
        thr.start()
        cst.play_media_url(stream.video_url, title=stream.video_title)
        click.echo("Serving local file, press Ctrl+C when done.")
        while thr.is_alive():
            time.sleep(1)

    elif stream.is_playlist:
        if stream.playlist_length == 0:
            cst.kill(idle_only=True)
            raise CattCliError("Playlist is empty.")
        click.echo("Casting remote playlist %s..." % video_url)
        if random_play:
            stream.set_playlist_entry(random.randrange(0, stream.playlist_length))
        else:
            try:
                if not stream.playlist_all_ids:
                    raise ValueError
                cst.play_playlist(stream.playlist_all_ids)
                return
            except (PlaybackError, ValueError):
                warning("Playlist playback not possible, playing first video.")
                stream.set_playlist_entry(0)
        click.echo("Playing %s on \"%s\"..." % (stream.playlist_entry_title, cst.cc_name))
        if cst.info_type == "url":
            cst.play_media_url(stream.playlist_entry_url,
                               title=stream.playlist_entry_title,
                               thumb=stream.playlist_entry_thumbnail)
        elif cst.info_type == "id":
            cst.play_media_id(stream.playlist_entry_id)

    else:
        click.echo("Casting remote file %s..." % video_url)
        click.echo("Playing %s on \"%s\"..." % (stream.video_title, cst.cc_name))
        if cst.info_type == "url":
            cst.play_media_url(stream.video_url, title=stream.video_title,
                               thumb=stream.video_thumbnail)
        elif cst.info_type == "id":
            cst.play_media_id(stream.video_id)


@cli.command(short_help="Add a video to the queue.")
@click.argument("video_url", callback=process_url)
@click.pass_obj
def add(settings, video_url):
    cst, stream = setup_cast(settings["device"], video_url=video_url, prep="control")
    if (cst.name != "default" and cst.name != stream.extractor) or not stream.is_video:
        raise CattCliError("This url cannot be added to the queue.")
    cst.add(stream.video_id)


@cli.command(short_help="Pause a video.")
@click.pass_obj
def pause(settings):
    cst = setup_cast(settings["device"], prep="control")
    cst.pause()


@cli.command(short_help="Resume a video after it has been paused.")
@click.pass_obj
def play(settings):
    cst = setup_cast(settings["device"], prep="control")
    cst.play()


@cli.command(short_help="Stop playing.")
@click.pass_obj
def stop(settings):
    cst = setup_cast(settings["device"])
    cst.kill()


@cli.command(short_help="Rewind a video by TIME duration.")
@click.argument("timedesc", type=CATT_TIME,
                required=False, default="30", metavar="TIME")
@click.pass_obj
def rewind(settings, timedesc):
    cst = setup_cast(settings["device"], prep="control")
    cst.rewind(timedesc)


@cli.command(short_help="Fastforward a video by TIME duration.")
@click.argument("timedesc", type=CATT_TIME,
                required=False, default="30", metavar="TIME")
@click.pass_obj
def ffwd(settings, timedesc):
    cst = setup_cast(settings["device"], prep="control")
    cst.ffwd(timedesc)


@cli.command(short_help="Seek the video to TIME position.")
@click.argument("timedesc", type=CATT_TIME, metavar="TIME")
@click.pass_obj
def seek(settings, timedesc):
    cst = setup_cast(settings["device"], prep="control")
    cst.seek(timedesc)


@cli.command(short_help="Skip to next video in queue (if any).")
@click.pass_obj
def skip(settings):
    cst = setup_cast(settings["device"], prep="control")
    cst.skip()


@cli.command(short_help="Set the volume to LVL [0-100].")
@click.argument("level", type=click.IntRange(0, 100), metavar="LVL")
@click.pass_obj
def volume(settings, level):
    cst = setup_cast(settings["device"])
    cst.volume(level / 100.0)


@cli.command(short_help="Turn up volume by a DELTA increment.")
@click.argument("delta", type=click.IntRange(1, 100),
                required=False, default=10, metavar="DELTA")
@click.pass_obj
def volumeup(settings, delta):
    cst = setup_cast(settings["device"])
    cst.volumeup(delta / 100.0)


@cli.command(short_help="Turn down volume by a DELTA increment.")
@click.argument("delta", type=click.IntRange(1, 100),
                required=False, default=10, metavar="DELTA")
@click.pass_obj
def volumedown(settings, delta):
    cst = setup_cast(settings["device"])
    cst.volumedown(delta / 100.0)


@cli.command(short_help="Show some information about the currently-playing video.")
@click.pass_obj
def status(settings):
    cst = setup_cast(settings["device"], prep="control")
    print_status(cst.cast_info)


@cli.command(short_help="Show complete information about the currently-playing video.")
@click.pass_obj
def info(settings):
    cst = setup_cast(settings["device"], prep="control")
    for (key, value) in cst.info.items():
        click.echo("%s: %s" % (key, value))


@cli.command(short_help="Scan the local network and show all Chromecasts and their IPs.")
def scan():
    click.echo("Scanning Chromecasts...")
    for device in get_chromecasts():
        click.echo("{0.host} - {0.device.friendly_name} - {0.device.manufacturer} {0.device.model_name}".format(device))


@cli.command(short_help="Save the current state of the Chromecast for later use.")
@click.argument("path",
                type=click.Path(writable=True), callback=process_path, required=False)
@click.pass_obj
def save(settings, path):
    cst = setup_cast(settings["device"], prep="control")
    if not cst.save_capability or cst.is_streaming_local_file:
        raise CattCliError("Saving state of this kind of content is not supported.")
    elif cst.save_capability == "partial":
        warning("Please be advised that playlist data will not be saved.")

    print_status(cst.media_info)
    if path and path.exists():
        click.confirm("File already exists. Overwrite?", abort=True)
    click.echo("Saving...")
    state = CastState(path or STATE_PATH, create_dir=True if not path else False)
    state.set_data(cst.cc_name, {"controller": cst.name, "data": cst.media_info})


@cli.command(short_help="Return Chromecast to saved state.")
@click.argument("path",
                type=click.Path(exists=True), callback=process_path, required=False)
@click.pass_obj
def restore(settings, path):
    cst = setup_cast(settings["device"])
    state = CastState(path or STATE_PATH)
    try:
        data = state.get_data(cst.cc_name)
    except StateFileError:
        raise CattCliError("The chosen file is not a valid save file.")
    if not data:
        raise CattCliError("No save data found for this device.")

    print_status(data["data"])
    click.echo("Restoring...")
    cst = setup_cast(settings["device"], prep="app", controller=data["controller"])
    cst.restore(data["data"])


def print_status(status):
    if status.get("title"):
        click.echo("Title: %s" % status["title"])

    if status.get("current_time"):
        current = human_time(status["current_time"])
        if status.get("duration"):
            duration = human_time(status["duration"])
            remaining = human_time(status["remaining"])
            click.echo("Time: %s / %s (%s%%)" % (current, duration, status["progress"]))
            click.echo("Remaining time: %s" % remaining)
        else:
            click.echo("Time: %s" % current)

    if status.get("player_state"):
        click.echo("State: %s" % status["player_state"])

    if status.get("volume_level"):
        click.echo("Volume: %s" % status["volume_level"])


def writeconfig(settings):
    try:
        CONFIG_DIR.mkdir()
    except FileExistsError:
        pass

    # Put all the standalone options from the settings into an "options" key.
    old_conf = readconfig()
    conf = {"options": settings}
    conf["aliases"] = old_conf["aliases"]

    # Convert the conf dict into a ConfigParser instance.
    config = configparser.ConfigParser()

    for section, options in conf.items():
        config.add_section(section)
        for option, value in options.items():
            config.set(section, option, value)

    with CONFIG_PATH.open("w") as configfile:
        config.write(configfile)


def readconfig():
    """
    Read the configuration from the config file.

    Returns a dictionary of the form:
        {"option": "value",
         "aliases": {"device1": "device_name"}}
    """
    config = configparser.ConfigParser()
    config.read(CONFIG_PATH)
    conf_dict = {section: dict(config.items(section)) for section in config.sections()}

    conf = conf_dict.get("options", {})
    conf["aliases"] = conf_dict.get("aliases", {})

    return conf


def main():
    return cli(obj={}, default_map=readconfig())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from catt.controllers import Cache
from catt.stream_info import StreamInfo


class TestThings(unittest.TestCase):
    def test_stream_info_youtube_video(self):
        stream = StreamInfo("https://www.youtube.com/watch?v=VZMfhtKa-wo")
        self.assertIn("https://", stream.video_url)
        self.assertEqual(stream.video_id, "VZMfhtKa-wo")
        self.assertTrue(stream.is_video)
        self.assertEqual(stream.extractor, "youtube")

    def test_stream_info_youtube_playlist(self):
        stream = StreamInfo("https://www.youtube.com/playlist?list=PL9Z0stL3aRykWNoVQW96JFIkelka_93Sc")
        self.assertIsNone(stream.video_url)
        self.assertEqual(stream.playlist_id, "PL9Z0stL3aRykWNoVQW96JFIkelka_93Sc")
        self.assertTrue(stream.is_playlist)
        self.assertEqual(stream.extractor, "youtube")

    def test_stream_info_other_video(self):
        stream = StreamInfo("http://www.dailymotion.com/video/x6fotne")
        self.assertIn("https://", stream.video_url)
        self.assertEqual(stream.video_id, "x6fotne")
        self.assertTrue(stream.is_video)
        self.assertEqual(stream.extractor, "dailymotion")

    def test_cache(self):
        cache = Cache()
        cache.set_data("key", "value")
        self.assertEqual(cache.get_data("key"), "value")

        cache.clear()
        cache = Cache()
        self.assertEqual(cache.get_data("key"), None)
        cache.clear()


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from catt import __version__
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    "youtube-dl>=2017.3.15",
    "PyChromecast>=2.0.0",
    "Click>=5.0",
]

test_requirements = [
]

setup(
    name='catt',
    version=__version__,
    description="Cast All The Things allows you to send videos from many, many online sources to your Chromecast.",
    long_description=readme,
    author="Stavros Korokithakis",
    author_email='hi@stavros.io',
    url='https://github.com/skorokithakis/catt',
    packages=[
        'catt',
    ],
    package_dir={'catt':
                 'catt'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords="chromecast cast catt cast_all_the_things",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    entry_points={
        'console_scripts': ['catt=catt.cli:main'],
    },
)



import json
import tempfile
import threading
from pathlib import Path

import pychromecast
from click import ClickException, echo

from .stream_info import StreamInfo
from .util import warning
from .youtube import YouTubeController


APP_INFO = [{"app_name": "youtube", "app_id": "233637DE", "supported_device_types": ["cast"]}]
DEFAULT_APP = {"app_name": "default", "app_id": "CC1AD845"}
BACKDROP_APP_ID = "E8C28D3C"


def get_chromecasts():
    devices = pychromecast.get_chromecasts()

    if not devices:
        raise CattCastError("No devices found.")

    devices.sort(key=lambda cc: cc.name)
    # We need to ensure that all Chromecast objects contain DIAL info.
    return [pychromecast.Chromecast(cc.host) for cc in devices]


def get_chromecast(device_name):
    devices = get_chromecasts()

    if device_name:
        try:
            return next(cc for cc in devices if cc.name == device_name)
        except StopIteration:
            raise CattCastError("Specified device \"%s\" not found." % device_name)
    else:
        return devices[0]


def setup_cast(device_name, video_url=None, prep=None, controller=None):
    """
    Prepares selected chromecast and/or media file.

    :param device_name: Friendly name of chromecast device to use.
    :type device_name: str or NoneType
    :param video_url: If supplied, setup_cast will try to exctract a media url
                      from this, for playback or queing.
    :type video_url: str
    :param prep: If prep = "app", video_url, if supplied, is meant for playback.
                 The relevant chromecast app is started during initialization
                 of the CastController object.
                 If prep = "control", video_url, if supplied, is meant for
                 queing. The state of the selected chromecast is determined
                 during initialization of the CastController object.
                 If prep = None, no preparation is done. Should only be used
                 if the desired action can be carried out regardless of the
                 state of the chromecast (like volume adjustment).
    :type prep: str
    :param controller: If supplied, the normal logic for determining the appropriate
                       controller is bypassed, and the one specified here is
                       returned instead.
    :type controller: str
    :returns: controllers.DefaultCastController or controllers.YoutubeCastController,
              and stream_info.StreamInfo if video_url is supplied.
    """

    cache = Cache()
    cached_ip = cache.get_data(device_name)
    stream = None

    try:
        if not cached_ip:
            raise ValueError
        cast = pychromecast.Chromecast(cached_ip)
    except (pychromecast.error.ChromecastConnectionError, ValueError):
        cast = get_chromecast(device_name)
        cache.set_data(cast.name, cast.host)
    cast.wait()

    if video_url:
        cc_info = (cast.device.manufacturer, cast.model_name)
        stream = StreamInfo(video_url, model=cc_info, host=cast.host)

    if controller:
        if controller == "default":
            app = DEFAULT_APP
        else:
            app = next(a for a in APP_INFO if a["app_name"] == controller)
    elif stream and prep == "app":
        if stream.is_local_file:
            app = DEFAULT_APP
        else:
            try:
                app = next(a for a in APP_INFO if a["app_name"] == stream.extractor)
            except StopIteration:
                app = DEFAULT_APP
    else:
        try:
            app = next(a for a in APP_INFO if a["app_id"] == cast.app_id)
        except StopIteration:
            app = DEFAULT_APP

    if (not controller and app["app_name"] != "default" and
            cast.cast_type not in app["supported_device_types"]):
        if stream:
            warning("The %s app is not available for this device." % app["app_name"].capitalize())
        app = DEFAULT_APP

    if app["app_name"] == "youtube":
        controller = YoutubeCastController(cast, app["app_name"], app["app_id"], prep=prep)
    else:
        controller = DefaultCastController(cast, app["app_name"], app["app_id"], prep=prep)
    return (controller, stream) if stream else controller


def catch_namespace_error(func):
    """
    Use this decorator for methods in CastController subclasses where the intended
    action is dependent on the chromecast being in a particular state (such as not
    buffering). If the cc app is then interrupted while catt is waiting for this state,
    we fail in a nice way.
    """

    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except pychromecast.error.UnsupportedNamespace:
            raise CattCastError("Chromecast app operation was interrupted.")
    return wrapper


class CattCastError(ClickException):
    pass


class StateFileError(Exception):
    pass


class PlaybackError(Exception):
    pass


class CattStore:
    def __init__(self, store_path):
        self.store_path = store_path

    def _create_store_dir(self):
        try:
            self.store_path.parent.mkdir()
        except FileExistsError:
            pass

    def _read_store(self):
        with self.store_path.open() as store:
            return json.load(store)

    def _write_store(self, data):
        with self.store_path.open("w") as store:
            json.dump(data, store)

    def get_data(self, *args):
        raise NotImplementedError

    def set_data(self, name, value):
        data = self._read_store()
        data[name] = value
        self._write_store(data)

    def clear(self):
        try:
            self.store_path.unlink()
            self.store_path.parent.rmdir()
        except FileNotFoundError:
            pass


class Cache(CattStore):
    def __init__(self):
        cache_path = Path(tempfile.gettempdir(), "catt_cache", "chromecast_hosts")
        super(Cache, self).__init__(cache_path)
        self._create_store_dir()

        if not self.store_path.exists():
            devices = pychromecast.get_chromecasts()
            self._write_store({d.name: d.host for d in devices})

    def get_data(self, name):
        data = self._read_store()
        # In the case that cache has been initialized with no cc's on the
        # network, we need to ensure auto-discovery.
        if not data:
            return None
        # When the user does not specify a device, we need to make an attempt
        # to consistently return the same IP, thus the alphabetical sorting.
        if not name:
            return data[min(data, key=str)]
        return data.get(name)


class CastState(CattStore):
    def __init__(self, state_path, create_dir=False):
        super(CastState, self).__init__(state_path)
        if create_dir:
            self._create_store_dir()
        if not self.store_path.exists():
            self._write_store({})

    def get_data(self, name):
        try:
            data = self._read_store()
            if set(next(iter(data.values())).keys()) != set(["controller", "data"]):
                raise ValueError
        except (json.decoder.JSONDecodeError,
                ValueError, StopIteration, AttributeError):
            raise StateFileError
        return data.get(name)


class CastStatusListener:
    def __init__(self, app_id, active_app_id):
        self.app_id = app_id
        self.app_ready = threading.Event()
        if app_id == active_app_id:
            self.app_ready.set()

    def new_cast_status(self, status):
        if status.app_id == self.app_id:
            self.app_ready.set()
        else:
            self.app_ready.clear()


class MediaStatusListener:
    def __init__(self, state):
        self.not_buffering = threading.Event()
        self.playing = threading.Event()
        if state != "BUFFERING":
            self.not_buffering.set()
        elif state == "PLAYING":
            self.playing.set()

    def new_media_status(self, status):
        if status.player_state == "BUFFERING":
            self.not_buffering.clear()
            self.playing.clear()
        elif status.player_state == "PLAYING":
            self.not_buffering.set()
            self.playing.set()
        else:
            self.not_buffering.set()
            self.playing.clear()


class CastController:
    def __init__(self, cast, name, app_id, prep=None):
        self._cast = cast
        self.name = name
        self.info_type = None
        self.save_capability = None

        self._cast_listener = CastStatusListener(app_id, self._cast.app_id)
        self._cast.register_status_listener(self._cast_listener)
        self._media_listener = MediaStatusListener(self._cast.media_controller.status.player_state)
        self._cast.media_controller.register_status_listener(self._media_listener)

        try:
            self._cast.register_handler(self._controller)
        except AttributeError:
            self._controller = self._cast.media_controller

        if prep == "app":
            self._prep_app()
        elif prep == "control":
            self._prep_control()

    @property
    def cc_name(self):
        return self._cast.device.friendly_name

    @property
    def info(self):
        status = self._cast.media_controller.status.__dict__
        # Values in media_controller.status for the keys "volume_level" and "volume_muted"
        # are always the same, regardless of actual state, so we discard those by
        # overwriting them with the values from system status.
        status.update(self._cast.status._asdict())
        return status

    @property
    def media_info(self):
        status = self._cast.media_controller.status
        return {"title": status.title,
                "content_id": status.content_id,
                "current_time": status.current_time if self._is_seekable else None,
                "thumb": status.images[0].url if status.images else None}

    @property
    def cast_info(self):
        status = self._cast.media_controller.status
        cinfo = self.media_info

        if self._is_seekable:
            duration, current = status.duration, status.current_time
            remaining = duration - current
            progress = int((1.0 * current / duration) * 100)
            cinfo.update({"duration": duration,
                          "remaining": remaining, "progress": progress})

        cinfo.update({"player_state": status.player_state,
                      "volume_level": str(int(round(self._cast.status.volume_level, 2) * 100))})
        return cinfo

    @property
    def is_streaming_local_file(self):
        status = self._cast.media_controller.status
        return True if status.content_id.endswith("?loaded_from_catt") else False

    @property
    def _is_seekable(self):
        status = self._cast.media_controller.status
        return True if (status.duration and
                        status.stream_type == "BUFFERED") else False

    def _prep_app(self):
        """Make shure desired chromecast app is running."""

        if not self._cast_listener.app_ready.is_set():
            self._cast.start_app(self._cast_listener.app_id)
            self._cast_listener.app_ready.wait()

    def _prep_control(self):
        """Make shure chromecast is in an active state."""

        if self._cast.app_id == BACKDROP_APP_ID or not self._cast.app_id:
            raise CattCastError("Chromecast is inactive.")
        self._cast.media_controller.block_until_active(1.0)
        if self._cast.media_controller.status.player_state in ["UNKNOWN", "IDLE"]:
            raise CattCastError("Nothing is currently playing.")

    def play_media_url(self, video_url, **kwargs):
        """
        CastController subclasses need to implement
        either play_media_url or play_media_id
        """

        raise NotImplementedError

    def play_media_id(self, video_id):
        """
        CastController subclasses need to implement
        either play_media_url or play_media_id
        """

        raise NotImplementedError

    def play_playlist(self, playlist_id):
        raise PlaybackError

    def play(self):
        self._cast.media_controller.play()

    def pause(self):
        self._cast.media_controller.pause()

    def seek(self, seconds):
        if self._is_seekable:
            self._cast.media_controller.seek(seconds)
        else:
            raise CattCastError("Stream is not seekable.")

    def rewind(self, seconds):
        pos = self._cast.media_controller.status.current_time
        self.seek(pos - seconds)

    def ffwd(self, seconds):
        pos = self._cast.media_controller.status.current_time
        self.seek(pos + seconds)

    def skip(self):
        if self._is_seekable:
            self._cast.media_controller.skip()
        else:
            raise CattCastError("Stream is not skippable.")

    def volume(self, level):
        self._cast.set_volume(level)

    def volumeup(self, delta):
        self._cast.volume_up(delta)

    def volumedown(self, delta):
        self._cast.volume_down(delta)

    def kill(self, idle_only=False):
        """
        Kills current Chromecast session.

        :param idle_only: If set, session is only killed if the active Chromecast app
                          is idle. Use to avoid killing an active streaming session
                          when catt fails with certain invalid actions (such as trying
                          to cast an empty playlist).
        :type idle_only: bool
        """

        if (idle_only and
                self._cast.media_controller.status.player_state not in ["UNKNOWN", "IDLE"]):
            return
        self._cast.quit_app()

    def restore(self, data):
        """
        Recreates Chromecast state from save data.
        Subclasses can implement this if its possible to recreate
        a session from save data.
        """

        raise NotImplementedError

    def _not_supported(self):
        self.kill(idle_only=True)
        raise CattCastError("This action is not supported by the %s controller." % self.name.capitalize())

    def add(self, video_id):
        self._not_supported()


class DefaultCastController(CastController):
    def __init__(self, cast, name, app_id, prep=None):
        super(DefaultCastController, self).__init__(cast, name, app_id, prep=prep)
        self.info_type = "url"
        self.save_capability = "complete" if (self._is_seekable and
                                              self._cast.app_id == DEFAULT_APP["app_id"]) else None

    def play_media_url(self, video_url, **kwargs):
        self._controller.play_media(video_url, "video/mp4",
                                    current_time=kwargs.get("current_time"),
                                    title=kwargs.get("title"), thumb=kwargs.get("thumb"))
        self._controller.block_until_active()

    def restore(self, data):
        self.play_media_url(data["content_id"], current_time=data["current_time"],
                            title=data["title"], thumb=data["thumb"])


class YoutubeCastController(CastController):
    def __init__(self, cast, name, app_id, prep=None):
        self._controller = YouTubeController()
        super(YoutubeCastController, self).__init__(cast, name, app_id, prep=prep)
        self.info_type = "id"
        self.save_capability = "partial"

    # The controller's start_new_session method needs a video id.
    def _prep_yt(self, video_id):
        if not self._controller.in_session:
            self._controller.start_new_session(video_id)

    def play_media_id(self, video_id):
        self._prep_yt(video_id)
        self._controller.play_video(video_id)

    def play_playlist(self, playlist):
        self.play_media_id(playlist[0])
        if len(playlist) > 1:
            for video_id in playlist[1:]:
                self.add(video_id)

    @catch_namespace_error
    def add(self, video_id):
        echo("Adding video id \"%s\" to the queue." % video_id)
        self._prep_yt(video_id)
        # You can't add videos to the queue while the app is buffering.
        self._media_listener.not_buffering.wait()
        self._controller.add_to_queue(video_id)

    @catch_namespace_error
    def restore(self, data):
        self.play_media_id(data["content_id"])
        self._media_listener.playing.wait()
        self.seek(data["current_time"])



import socketserver
import time
import traceback
from http.server import BaseHTTPRequestHandler
from pathlib import Path


def serve_file(filename, address="", port=45114):
    class FileHandler(BaseHTTPRequestHandler):
        def do_GET(self):  # noqa
            try:
                mediapath = Path(filename)
                length = mediapath.stat().st_size
                mtime = mediapath.stat().st_mtime
                mediafile = open(str(mediapath), "rb")

                self.send_response(200)
                self.send_header("Content-type", "video/mp4")
                self.send_header("Content-Length", length)
                self.send_header(
                    "Last-Modified",
                    time.strftime(
                        "%a %d %b %Y %H:%M:%S GMT",
                        time.localtime(mtime)
                    )
                )
                self.end_headers()

                while True:
                    data = mediafile.read(100 * 1024)

                    if not data:
                        break
                    self.wfile.write(data)
            except:  # noqa
                traceback.print_exc()

            mediafile.close()

    httpd = socketserver.TCPServer((address, port), FileHandler)
    httpd.serve_forever()
    httpd.server_close()



"""
Controller to interface with the YouTube-app.
"""

import re
import threading
try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

import requests
from pychromecast.controllers import BaseController

YOUTUBE_BASE_URL = "https://www.youtube.com/"
YOUTUBE_WATCH_VIDEO_URL = YOUTUBE_BASE_URL + "watch?v="

# id param is const(YouTube sets it as random xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx so it should be fine).
RANDOM_ID = "12345678-9ABC-4DEF-0123-0123456789AB"
VIDEO_ID_PARAM = '%7B%22videoId%22%3A%22{video_id}%22%2C%22currentTime%22%3A5%2C%22currentIndex%22%3A0%7D'
TERMINATE_PARAM = "terminate"

REQUEST_URL_SET_PLAYLIST = YOUTUBE_BASE_URL + "api/lounge/bc/bind?"
BASE_REQUEST_PARAMS = {"device": "REMOTE_CONTROL", "id": RANDOM_ID, "name": "Desktop&app=youtube-desktop",
                       "mdx-version": 3, "loungeIdToken": None, "VER": 8, "v": 2, "t": 1, "ui": 1, "RID": 75956,
                       "CVER": 1}

SET_PLAYLIST_METHOD = {"method": "setPlaylist", "params": VIDEO_ID_PARAM, "TYPE": None}
REQUEST_PARAMS_SET_PLAYLIST = dict(**dict(BASE_REQUEST_PARAMS, **SET_PLAYLIST_METHOD))

REQUEST_DATA_SET_PLAYLIST = "count=0"
REQUEST_DATA_ADD_TO_PLAYLIST = "count=1&ofs=%d&req0__sc=addVideo&req0_videoId=%s"
REQUEST_DATA_REMOVE_FROM_PLAYLIST = "count=1&ofs=%d&req0__sc=removeVideo&req0_videoId=%s"
REQUEST_DATA_CLEAR_PLAYLIST = "count=1&ofs=%d&req0__sc=clearPlaylist"

REQUEST_URL_LOUNGE_TOKEN = YOUTUBE_BASE_URL + "api/lounge/pairing/get_lounge_token_batch"
REQUEST_DATA_LOUNGE_TOKEN = "screen_ids={screenId}&session_token={XSRFToken}"

YOUTUBE_SESSION_TOKEN_REGEX = 'XSRF_TOKEN\W*(.*)="'
SID_REGEX = '"c","(.*?)",\"'
PLAYLIST_ID_REGEX = 'listId":"(.*?)"'
FIRST_VIDEO_ID_REGEX = 'firstVideoId":"(.*?)"'
GSESSION_ID_REGEX = '"S","(.*?)"]'
NOW_PLAYING_REGEX = 'videoId":"(.*?)"'

EXPIRED_LOUNGE_ID_RESPONSE_CONTENT = "Expired lounge id token"

MEDIA_NAMESPACE = "urn:x-cast:com.google.cast.media"
MESSAGE_TYPE = "type"
TYPE_GET_SCREEN_ID = "getMdxSessionStatus"
TYPE_STATUS = "mdxSessionStatus"
ATTR_SCREEN_ID = "screenId"
TYPE_PLAY = "PLAY"
TYPE_PAUSE = "PAUSE"
TYPE_STOP = "STOP"


class YoutubeSessionError(Exception):
    pass


class YoutubeControllerError(Exception):
    pass


class YouTubeController(BaseController):
    """ Controller to interact with Youtube."""

    def __init__(self):
        super(YouTubeController, self).__init__(
            "urn:x-cast:com.google.youtube.mdx", "233637DE")

        self._xsrf_token = None
        self._lounge_token = None
        self._gsession_id = None
        self._sid = None
        self._ofs = 0
        self._first_video = None
        self._playlist_id = None
        self.screen_id = None
        self.video_id = None
        self.playlist = None
        self._now_playing = None
        self.status_update_event = threading.Event()

    @property
    def video_url(self):
        """Returns the base watch video url with the current video_id"""
        video = self._now_playing or self.video_id
        return YOUTUBE_WATCH_VIDEO_URL + video

    @property
    def status(self):
        """ Returns the media_controller status handler when Youtube app is launched."""
        if self.is_active:
            return self._socket_client.media_controller.status
        else:
            return None

    @property
    def in_session(self):
        """ Returns True if session params are not None."""
        if self._gsession_id and self._sid and self._lounge_token:
            return True
        else:
            return False

    def _do_post(self, url, data, params=None, referer=None):
        """
        Does all post requests.
        will raise if response is not 200-ok
        :param url:(str)the request url
        :param data:(str) the request body
        :param params:(dict) the request urlparams
        :param referer:(str) the referer. default is the video url that started the session.
        :return: the response
        """
        headers = {
            "Origin": YOUTUBE_BASE_URL,
            "Content-Type": "application/x-www-form-urlencoded",
            "Referer": (referer or self.video_url)
        }
        response = requests.post(url, headers=headers, data=data, params=params)
        response.raise_for_status()
        return response

    def update_screen_id(self):
        """
        Sends a getMdxSessionStatus to get the screen id and waits for response.
        This function is blocking but if connected we should always get a response
        (send message will launch app if it is not running).
        """
        self.status_update_event.clear()
        self.send_message({MESSAGE_TYPE: TYPE_GET_SCREEN_ID})
        self.status_update_event.wait()
        self.status_update_event.clear()

    def _get_xsrf_token(self):
        """
        Get the xsrf_token used as the session token.
        video_id must be initialized.
        Sets the session token(xsrf_token).
        """
        if not self.video_id:
            raise ValueError("Cant start a session without the video_id.")
        response = requests.get(self.video_url)
        response.raise_for_status()
        token = re.search(YOUTUBE_SESSION_TOKEN_REGEX, str(response.content))
        if not token:
            raise YoutubeSessionError("Could not fetch the xsrf token")
        self._xsrf_token = token.group(1)

    def _get_lounge_id(self):
        """
        Gets the lounge_token.
        session_token(xsrf_token) and screenId must be initialized.
        Sets the lounge token.
        """
        if not self.screen_id:
            raise ValueError("Screen id is None. update_screen_id must be called.")
        if not self._xsrf_token:
            raise ValueError("xsrf token is None. Get xsrf token must be called.")
        data = REQUEST_DATA_LOUNGE_TOKEN.format(screenId=self.screen_id, XSRFToken=self._xsrf_token)
        response = self._do_post(REQUEST_URL_LOUNGE_TOKEN, data=data)
        if response.status_code == 401:
            # Screen id is not None and it is updated with a message from the Chromecast.
            #  It is very unlikely that screen_id caused the problem.
            raise YoutubeSessionError("Could not get lounge id. XSRF token has expired or is not valid.")
        response.raise_for_status()
        try:
            lounge_token = response.json()["screens"][0]["loungeToken"]
        except JSONDecodeError:
            raise YoutubeSessionError("Could not get lounge id. XSRF token has expired or not valid.")
        self._lounge_token = lounge_token

    def _set_playlist(self):
        """
        Sends a POST to start the session.
        Uses loung_token and video id as parameters.
        Sets session SID and gsessionid on success.
        """
        if not self.video_id:
            raise ValueError("Can't start a session without the video_id.")
        if not self._lounge_token:
            raise ValueError("lounge token is None. _get_lounge_token must be called")
        url_params = REQUEST_PARAMS_SET_PLAYLIST.copy()
        url_params['loungeIdToken'] = self._lounge_token
        url_params['params'] = VIDEO_ID_PARAM.format(video_id=self.video_id)
        response = self._do_post(REQUEST_URL_SET_PLAYLIST, data=REQUEST_DATA_SET_PLAYLIST, params=url_params)
        content = str(response.content)
        if response.status_code == 401 and content.find(EXPIRED_LOUNGE_ID_RESPONSE_CONTENT) != -1:
            raise YoutubeSessionError("The lounge token expired.")
        response.raise_for_status()
        if not self.in_session:
            self._extract_session_parameters(content)

    def _update_session_parameters(self):
        """
        Sends a POST with no playlist parameters.
        Gets the playlist id, SID, gsession id.
        First video(the playlist base video) and now playing are also returned if  playlist is initialized.
        """
        url_params = BASE_REQUEST_PARAMS.copy()
        url_params['loungeIdToken'] = self._lounge_token
        response = self._do_post(REQUEST_URL_SET_PLAYLIST, data='', params=url_params)
        self._extract_session_parameters(str(response.content))
        return response

    def _extract_session_parameters(self, response_packet_content):
        """
        Extracts the playlist id, SID, gsession id, first video(the playlist base video)
        and now playing from a session response.
        :param response_packet_content: (str) the response packet content
        """
        content = response_packet_content
        playlist_id = re.search(PLAYLIST_ID_REGEX, content)
        sid = re.search(SID_REGEX, content)
        gsession = re.search(GSESSION_ID_REGEX, content)
        first_video = re.search(FIRST_VIDEO_ID_REGEX, content)
        now_playing = re.search(NOW_PLAYING_REGEX, content)
        if not (sid and gsession and playlist_id):
            raise YoutubeSessionError("Could not parse session parameters.")
        self._sid = sid.group(1)
        self._gsession_id = gsession.group(1)
        self._playlist_id = playlist_id.group(1)
        if first_video:
            self._first_video = first_video.group(1)
        else:
            self._first_video = None
        if now_playing:
            self._now_playing = now_playing.group(1)
        else:
            self._now_playing = None

    def _manage_playlist(self, data, referer=None, **kwargs):
        """
        Manages all request to an existing session.
        _gsession_id, _sid, video_id and _lounge_token must be initialized.
        :param data: data of the request
        :param video_id: video id in the request
        :param refer: used for the request heders referer field.video_url by default.
        """
        if not self._gsession_id:
            raise ValueError("gsession must be initialized to manage playlist")
        if not self._sid:
            raise ValueError("sid must be initialized to manage playlist")
        if not self.video_id:
            raise ValueError("video_id can't be empty")
        if self.in_session:
            self._update_session_parameters()
        param_video_id = self._first_video or self.video_id

        url_params = REQUEST_PARAMS_SET_PLAYLIST.copy()
        url_params["loungeIdToken"] = self._lounge_token
        url_params["params"] = VIDEO_ID_PARAM.format(video_id=param_video_id)
        url_params["gsessionid"] = self._gsession_id
        url_params["SID"] = self._sid
        for key in kwargs:
            if key in url_params:
                url_params[key] = kwargs[key]
        try:
            self._do_post(REQUEST_URL_SET_PLAYLIST, referer=referer, data=data, params=url_params)
        except requests.HTTPError:
            # Try to re-get session variables and post again.
            self._set_playlist()
            url_params["loungeIdToken"] = self._lounge_token
            url_params["params"] = VIDEO_ID_PARAM.format(video_id=self._first_video)
            url_params["gsessionid"] = self._gsession_id
            url_params["SID"] = self._sid
            self._ofs = 0
            self._do_post(REQUEST_URL_SET_PLAYLIST, referer=referer, data=data, params=url_params)

    def clear_playlist(self, terminate_session=False):
        """
        clears all tracks on queue without closing the session.
        terminate_session: close the existing session after clearing playlist.
        App closes after a few minutes idle so terminate session if idle for a few minutes.
        """
        self._ofs += 1
        self._manage_playlist(REQUEST_DATA_CLEAR_PLAYLIST % self._ofs)
        if terminate_session:
            self.terminate_session()
        self.playlist = None

    def terminate_session(self):
        """
        terminates the open lounge session.
        """
        try:
            self.clear_playlist()
            self._manage_playlist(data='', video_id=self.video_id, TYPE=TERMINATE_PARAM)
        except requests.RequestException:
            # Session has expired or not in sync.Clean session parameters anyway.
            pass
        self.screen_id = None
        self.video_id = None
        self._xsrf_token = None
        self._lounge_token = None
        self._gsession_id = None
        self._sid = None
        self._ofs = 0
        self.playlist = None
        self._first_video = None

    def receive_message(self, message, data):
        """ Called when a media message is received. """
        if data[MESSAGE_TYPE] == TYPE_STATUS:
            self._process_status(data.get("data"))

            return True

        else:
            return False

    def start_new_session(self, youtube_id):
        self.video_id = youtube_id
        self.update_screen_id()
        self._get_xsrf_token()
        self._get_lounge_id()
        self._update_session_parameters()

    def play_video(self, youtube_id):
        """
        Starts playing a video in the YouTube app.
        The youtube id is also a session identifier used in all requests for the session.
        :param youtube_id: The video id to play.
        """
        if not self.in_session:
            self.start_new_session(youtube_id)
        if self._first_video:
            self.clear_playlist()
        self._set_playlist()
        self._update_session_parameters()

    def add_to_queue(self, youtube_id):
        """
        Adds a video to the queue video will play after the currently playing video ends.
        If video is buffering it wil not be added!
        :param youtube_id: The video id to add to the queue
        """
        if not self.in_session:
            raise YoutubeSessionError('Session must be initialized to add to queue')
        if not self.playlist:
            self.playlist = [self.video_id]
        elif youtube_id in self.playlist:
            raise YoutubeControllerError("Video already in queue")
        self.update_screen_id()
        # if self.status.player_is_idle:
        #     raise YoutubeControllerError("Can't add to queue while video is idle")
        if self.status.player_state == "BUFFERING":
            raise YoutubeControllerError("Can't add to queue while video is buffering")
        self._ofs += 1
        self._manage_playlist(data=REQUEST_DATA_ADD_TO_PLAYLIST % (self._ofs, youtube_id))
        self.playlist.append(youtube_id)

    def _send_command(self, message, namespace=MEDIA_NAMESPACE):
        """
        Sends a message to a specific namespace.
        :param message:(dict) the message to sent to chromecast
        :param namespace:(str) the namespace to send the message to. default is media namespace.
        """
        self._socket_client.send_app_message(namespace, message)

    def play(self):
        self._send_command({MESSAGE_TYPE: TYPE_PLAY})

    def pause(self):
        self._send_command({MESSAGE_TYPE: TYPE_PAUSE})

    def stop(self, clear_queue=True):
        if clear_queue:
            self.clear_playlist()
        self._send_command({MESSAGE_TYPE: TYPE_STOP})

    def _process_status(self, status):
        """ Process latest status update. """
        self.screen_id = status.get(ATTR_SCREEN_ID)
        self.status_update_event.set()

    def tear_down(self):
        """ Called when controller is destroyed. """
        super(YouTubeController, self).tear_down()
        self.terminate_session()
